/* ccgo: go/go.cc
 * 
 * Copyright 2003 Chun-Chung Chen <cjj@u.washington.edu>
 * 
 * This program is free software; you can redistribute it and'or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 */
#include <go.hh>

using namespace go;
using namespace std;

// members of class Board
Board::Board(int s)
{
	z = s;
	b = new State[z * z];
}

Board::~Board()
{
	delete b;
}

Board::State Board::state(int l)
{
	return b[l];
}

void Board::state(int l, State s)
{
	b[l] = s;
}

vector<int> Board::neighbors(int l)
{
	vector<int> n;
	if (l % z > 0) n.push_back(l - 1);
	if (l % z < z - 1) n.push_back(l + 1);
	if (l / z > 0) n.push_back(l - z);
	if (l / z < z - 1) n.push_back(l + z);
	return n;
}

int Board::loc_range()
{
	return z * z;
}

// members of class Rule
Rule::Rule(Game & a) :
	g(a)
{
}

bool Rule::check(int l, vector<int> e)
{
	if (g.b->state(l) != Board::STATE_EMPTY) return false;
	if (kos.size() && kos.back() >= 0 && l == kos.back()) return false;
	return true;
}

void Rule::regis()
{
	Move * m = g.n->move;
	if (! m) return;

	// simple ko
	int t;
	if (m->loc < 0 || m->e.size() != 1) t = - 1; // pass
	else {
		vector<int> e;
		int r = g.b->loc_range();
		bool mk[r];
		for (int i = 0; i < r; i ++) mk[i] = false;
		Board::State s = g.b->state(m->loc);
		if (g.qi_group(m->loc, e, s, mk) == 1 && e.size() == 1) t = m->e.back();
		else t = - 1;
	}
	kos.push_back(t);
}

void Rule::unreg()
{
	kos.pop_back();
}

Node::Node() :
	move(0),
	up(0)
{
}

Node::~Node()
{
	while (down.size()) {
		delete down.back();
		down.pop_back();
	}
	if (move) delete move;
}

// members of class Game
Game::Game() :
	root(0),
	n(0),
	b(0),
	r(0)
{

}

Game::~Game()
{
}

Node * Game::new_node()
{
	return new Node;
}

void Game::set_board(Board * bd)
{
	b = bd;
}

void Game::set_rule(Rule * ru)
{
	r = ru;
}

void Game::init()
{
	if (root) delete root;
	root = new_node();
	n = root;
	if (! b) b = new Board(19);
	if (! r) r = new Rule(* this);
	turn = TURN_BLACK;
	num = 0;
}

void Game::add_move(int l)
{
	if (! (n || b)) return;
	vector<int> e;
	if (l >= 0) get_kill(l, e);
	if (! r->check(l, e)) return;
	if (n->move) {
		n->down.push_back(new_node());
		n->down.back()->up = n;
		n = n->down.back();
	}

	// set Move
	Move * m = new Move;
	m->loc = l;
	m->s = b->state(l);
	m->e = e;
	if (e.size()) m->es = b->state(e.back());
	n->move = m;

	// set board
	b->state(l, Board::State(turn));
	for (vector<int>::iterator i = e.begin(); i != e.end(); i ++) b->state(* i, Board::STATE_EMPTY);
	turn = Turn(3 - turn);
	num = num + 1;
	// register move
	r->regis();
}

void Game::up()
{
	if (! n->up) return;
	if (n->move) {
		r->unreg();
		// undo the move
		num = num - 1;
		turn = Turn(3 - turn);
		for (vector<int>::reverse_iterator i = n->move->e.rbegin(); i != n->move->e.rend(); i ++) b->state(* i, n->move->es);
		b->state(n->move->loc, n->move->s);
	}
	n = n->up;
}

void Game::down(int m)
{
	if (n->down.size() <= m) return;
	n = n->down[m];
	if (! n->move) return;
	b->state(n->move->loc, Board::State(turn));
	for (vector<int>::iterator i = n->move->e.begin(); i != n->move->e.end(); i ++) b->state(* i, Board::STATE_EMPTY);
	turn = Turn(3 - turn);
	num = num + 1;
}

void Game::get_kill(int l, vector<int> & k)
{
	int z = b->loc_range();
	bool m[z];
	Board::State s = Board::State(3 - turn); // check other stones around
	vector<int> n = b->neighbors(l); // checking the neighbors of l
	k.clear();
	int bn = 0; // number of killed stones
	while (n.size()) {
		for (int i = 0; i < z; i ++) m[i] = false;
		m[l] = true; // fill the point!
		int q = qi_group(n.back(), k, s, m);
		n.pop_back();
		for (vector<int>::iterator i = n.begin(); i != n.end(); i ++) if (m[* i]) {
			// remove checked neighbor from the list
			n.erase(i);
			i --;
		}
		if (q > 0) k.resize(bn); // group is alive, remove from k
		else bn = k.size();
	}
	if (bn > 0) return; // kill something

	// check for suicide
	for (int i = 0; i < z; i ++) m[i] = false;
	m[l] = true;
	k.push_back(l);
	int q = 0;
	n = b->neighbors(l);
	s = Board::State(turn); // check same stones around
	while (n.size()) {
		q += qi_group(n.back(), k, s, m);
		n.pop_back();
	}
	if (q) k.clear();
}

int Game::qi_group(int l, vector<int> & g, Board::State s, bool * m)
{
	if (m[l]) return 0;
	m[l] = true;
	if (b->state(l) == Board::STATE_EMPTY) return 1;
	if (b->state(l) != s) return 0;
	g.push_back(l);
	int q = 0;
	vector<int> n = b->neighbors(l);
	while (n.size()) {
		q += qi_group(n.back(), g, s, m);
		n.pop_back();
	}
	return q;
}
